from requests import Session

from .models import Cache


class YConverter:
    FIAT_BASE = "https://openexchangerates.org/api/latest.json"
    CRYPTO_BASE = "https://api.binance.com/api/v3/ticker/price"

    _cache: Cache
    _session: Session

    def __init__(self, api_key: str = "") -> None:
        self._cache = Cache.load()
        self._session = Session()
        if api_key:
            self._cache.api_key = api_key
        assert (
            self._cache.api_key
        ), "Get free API key from: https://openexchangerates.org/signup/free"

        if not self._cache.price_info:
            self.fetch_currencies()
            self._cache.save()

    def _fetch_fiat_currencies(self):
        for symbol, value in (
            self._session.get(f"{YConverter.FIAT_BASE}?app_id={self._cache.api_key}")
            .json()["rates"]
            .items()
        ):
            self._cache.cache(f"USD{symbol}", value, False)

    def _fetch_crypto_currencies(self):
        for data in self._session.get(YConverter.CRYPTO_BASE).json():
            self._cache.cache(data["symbol"], float(data["price"]), True)

    def fetch_currencies(self):
        self._fetch_fiat_currencies()
        self._fetch_crypto_currencies()

    def convert(self, amount: float, source: str, destination: str) -> float:
        def load_price(pair: str) -> float:
            if self._cache.is_fiat(pair):
                self._fetch_fiat_currencies()
                self._cache.save()
                return self._cache.get_price(pair)
            if self._cache.is_crypto(pair):
                self._fetch_crypto_currencies()
                self._cache.save()
                return self._cache.price_info[pair].value
            raise RuntimeError(
                f"{source} {destination} is not FIAT nor Crypto but it's cached."
            )

        pair = f"{source}{destination}".upper()
        if self._cache.contains(pair):
            if not self._cache.is_outdated(pair):
                return amount * self._cache.get_price(pair)
            else:
                return amount * load_price(pair)
        else:
            pair = f"{destination}{source}".upper()
            if self._cache.contains(pair):
                if not self._cache.is_outdated(pair):
                    return amount / self._cache.get_price(pair)
                else:
                    return amount / load_price(pair)

        if self._cache.is_fiat_symbol(source) and self._cache.is_fiat_symbol(
            destination
        ):
            prices: list[float] = []
            for pair in [f"USD{source}", f"USD{destination}"]:
                if not self._cache.is_outdated(pair):
                    prices.append(self._cache.get_price(pair))
                else:
                    prices.append(load_price(pair))
            return amount * prices[1] / prices[0]

        raise ValueError(
            f"{source} {destination} can't found in Fiat and Binance currencies"
        )



from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import Dict

from ruamel.yaml import YAML, yaml_object

yaml = YAML()


@yaml_object(yaml)
@dataclass
class PriceInfo:
    FIAT_CACHE_TIME = 10 * 60
    CRPYTO_CACHE_TIME = 10

    value: float
    is_crypto: bool

    timestamp: float = field(init=False)

    def __post_init__(self):
        self.timestamp = time()

    def is_outdated(self) -> bool:
        cache_time = (
            PriceInfo.FIAT_CACHE_TIME if self.is_crypto else PriceInfo.CRPYTO_CACHE_TIME
        )
        return time() - self.timestamp > cache_time


@yaml_object(yaml)
@dataclass
class Cache:
    PATH = Path.home() / "yconverter.yml"

    api_key: str = field(default="")
    price_info: Dict[str, PriceInfo] = field(default_factory=dict)

    @classmethod
    def load(cls) -> "Cache":
        return (
            yaml.load(Cache.PATH)
            if (Cache.PATH.exists() and Cache.PATH.read_text() != "")
            else cls()
        )

    def is_cached(self, pair: str) -> bool:
        pair = pair.upper()
        return pair in self.price_info and not self.price_info[pair].is_outdated()

    def is_fiat(self, pair: str) -> bool:
        pair = pair.upper()
        return pair in self.price_info and not self.price_info[pair].is_crypto

    def is_fiat_symbol(self, symbol: str) -> bool:
        return f"USD{symbol.upper()}" in self.price_info

    def is_crypto(self, pair: str) -> bool:
        pair = pair.upper()
        return pair in self.price_info and self.price_info[pair].is_crypto

    def cache(self, pair: str, value: float, is_crypto: bool):
        self.price_info[pair.upper()] = PriceInfo(value, is_crypto)

    def get_price(self, pair: str) -> float:
        return self.price_info[pair.upper()].value

    def is_outdated(self, pair: str) -> bool:
        return self.price_info[pair.upper()].is_outdated()

    def contains(self, pair: str) -> bool:
        return pair.upper() in self.price_info

    def save(self):
        yaml.dump(self, Cache.PATH)



from glob import glob
from os.path import basename, splitext

from setuptools import find_packages, setup

VERSION = "1.0.4"
README_PATH = "docs/README.md"

# test_requirements = ["behave", "behave-classy", "pytest"]

long_description = ""
with open(README_PATH, "r", encoding="utf-8") as file:
    long_description = file.read()

setup(
    name="yconverter",
    version=VERSION,
    license="Apache Software License 2.0",
    description="All fiat currency and crypto converter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Yunus Emre Ak",
    author_email="yemreak.com@gmail.com",
    url="https://github.com/yedhrab/YConverter",
    packages=find_packages(),
    py_modules=[splitext(basename(path))[0] for path in glob("*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Utilities",
    ],
    project_urls={
        "Changelog": "https://github.com/yedhrab/YConverter/blob/main/CHANGELOG.md",
        "Issue Tracker": "https://github.com/yedhrab/YConverter/issues",
    },
    keywords=[
        "cryptocurrencyconverter",
        "currencyconverter",
        "converter",
        "currecies",
        "crpytocurrencies",
        "cryptoconverter",
    ],
    python_requires=">=3.9",
    install_requires=["ruamel.yaml==0.17.21", "requests==2.28.2"],
    entry_points={"console_scripts": ["convert = yconverter.cli:main"]},
)
